# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import sys
import os
import os.path
import logging
import math

import cvxopt.base as cb

import logging
logging.basicConfig(level=logging.DEBUG,format='%(levelname)s %(message)s')

import CPX
import numpy as ny
import numpy.matlib as nyml
from cplex_const import *

from SIQP import SIQP

OPTIMAL = [CPX_STAT_OPTIMAL, CPXMIP_OPTIMAL, CPXMIP_OPTIMAL_TOL]
UNBOUNDED = [CPX_STAT_UNBOUNDED, CPXMIP_UNBOUNDED]
INFEASIBLE = [CPX_STAT_INFEASIBLE, CPXMIP_INFEASIBLE, CPXMIP_INForUNBD]

class FloatMatrix:
   
   TYPE = nyml.float64

   def __init__(self,list):
      self.data = nyml.mat(list, dtype=self.TYPE).T

class IntMatrix:
   
   TYPE = nyml.int32

   def __init__(self,list):
      self.data = nyml.mat(list, dtype=self.TYPE).T

class CharMatrix:
   
   TYPE = '|S1'

   def __init__(self,list):
      self.data = nyml.mat(list, dtype=self.TYPE).T

class SIQPSolver(SIQP):
   """
   This algorithm solves the problem 

   min 1/2 x'Px + q'x
   s.t. 
         Gx >= h
   """

   # Define types and constants
   Inf   = CPX_INFBOUND

   def __init__(self,fSize,numExamples,c,proto,run):
      SIQP.__init__(self,fSize,numExamples,c,run)
      self.numFeatures = fSize
      self.numExamples = numExamples
      self.numVariables = self.numFeatures + self.numExamples
      self.old_w = None
      self.old_objective_value = -(sys.maxint-1)

      self.addedConstraints = 0

      self.protocol = proto

      self.objsen = CPX_MIN

      # open CPLEX environment, set some parameters
      self.env = CPX.openCPLEX()

      CPX.setintparam(self.env, CPX_PARAM_SCRIND, CPX_ON) # print >> self.protocol, info to screen
      CPX.setintparam(self.env, CPX_PARAM_DATACHECK, CPX_ON)
      #CPX.setintparam(self.env, CPX_PARAM_QPMETHOD, 2)

      # create CPLEX problem, add objective and constraints to it
      self.lp = CPX.createprob(self.env, 'test1')

      self.obj = FloatMatrix([0.0]*self.numFeatures + [1.0]*self.numExamples).data
      cFactor = 1.0*self.C/self.numExamples
      self.obj =  cFactor * self.obj
      assert self.obj.shape == (self.numVariables,1)
      
      self.ub = FloatMatrix([self.Inf] * self.numVariables).data
      self.lb = FloatMatrix([-self.Inf] * self.numFeatures + [0.0]*self.numExamples).data

      # if we disable to intron length model we set the first boundaries of the
      # intron support points to 0.0
      #
      #import pdb
      #pdb.set_trace()

      if not run['enable_intron_length']:
         self.ub[0] = 0.0
         self.ub[1] = 0.0

         self.lb[0] = 0.0
         self.lb[1] = 0.0

      #pdb.set_trace()

      ############################

      self.matbeg = IntMatrix([1]*self.numVariables).data
      self.matbeg[0] = 0

      self.matcnt = IntMatrix([0]*self.numVariables).data
      self.matcnt[0] = 1

      self.matind = IntMatrix([0]).data
      self.matval = FloatMatrix([0.0]).data

      self.rhs = FloatMatrix([0.0]*self.numVariables).data
      self.sense = CharMatrix(['G']*self.numVariables).data

      self.numConstraints = 1

      CPX.copylp(self.env, self.lp, self.numVariables, self.numConstraints, self.objsen, 
               self.obj, self.rhs, self.sense,
               self.matbeg, self.matcnt, self.matind, self.matval, 
               self.lb, self.ub) 

      import pdb
      #pdb.set_trace()

      assert sum(self.P[self.numFeatures:,self.numFeatures:]) == 0.0, 'slack variables are regularized'

      self.qmatbeg, self.qmatcnt, self.qmatind, self.qmatval = self.cpx_matrix(self.P)
      CPX.copyquad(self.env, self.lp, self.qmatbeg, self.qmatcnt, self.qmatind, self.qmatval)

   def enforceMonotonicity(self, begin, end):
      """
      This method enforces monotonicity on the parameters indexed by begin to
      end.
      """
      assert -1 < begin < end < self.numFeatures

      for idx in xrange(begin,end):
         self.enforcePairwiseMonotonicity(idx,idx+1)

   def enforcePairwiseMonotonicity(self,pos1,pos2):
      energy_deltas = IntMatrix([0.0]*self.numVariables).data 

      energy_deltas[pos1] = -1.0
      energy_deltas[pos2] = 1.0

      numNewRows = 1
      numNewCols = 0
   
      row_rhs = FloatMatrix([0.0]).data
      row_sense = CharMatrix(['G']).data

      row_matbeg = IntMatrix([0]).data

      row_entries = [0.0]*self.numVariables
      indices = [0]*self.numVariables

      currentVal = 0.0
      numNonZeroElems = 0
      for idx in range(self.numFeatures):
         currentVal = energy_deltas[idx]
         if math.fabs(currentVal) > self.EPSILON:
            row_entries[numNonZeroElems] = currentVal
            indices[numNonZeroElems] = idx
            numNonZeroElems += 1

      row_matind = IntMatrix([0]*(numNonZeroElems+1)).data
      row_matval = FloatMatrix([0.0]*(numNonZeroElems+1)).data

      for pos in range(numNonZeroElems):
         row_matind[pos] = indices[pos]
         row_matval[pos] = row_entries[pos]

      status = CPX.addrows (self.env, self.lp, numNewCols, numNewRows, numNonZeroElems+1, row_rhs,
                                  row_sense, row_matbeg, row_matind, row_matval)
      
      assert status == 0, 'Error ocurred while adding constraint.'

      self.addedConstraints += 1

      return True

   def addConstraint(self, energy_deltas, example_idx):
   #def addConstraint(self, energy_deltas, loss, example_idx, useMarginRescaling):
      """
      This method adds one contraint to the optimization problem.
      """
      self.lastDelta = energy_deltas
      self.lastIndex = example_idx

      assert -1 < example_idx and example_idx < self.numExamples

      if self.old_w != None:
         print >> self.protocol, 'Scalar prod. contrib is %f'%((self.old_w[0:self.numFeatures].T*energy_deltas)[0])

      numNewRows = 1
      numNewCols = 0
   
      row_rhs = FloatMatrix([1.0]).data
      row_sense = CharMatrix(['G']).data

      row_matbeg = IntMatrix([0]).data

      row_entries = [0.0]*self.numVariables
      indices = [0]*self.numVariables

      currentVal = 0.0
      numNonZeroElems = 0
      for idx in range(self.numFeatures):
         currentVal = energy_deltas[idx]
         if math.fabs(currentVal) > self.EPSILON:
            row_entries[numNonZeroElems] = currentVal
            indices[numNonZeroElems] = idx
            numNonZeroElems += 1

      row_matind = IntMatrix([0]*(numNonZeroElems+1)).data
      row_matval = FloatMatrix([0.0]*(numNonZeroElems+1)).data

      for pos in range(numNonZeroElems):
         row_matind[pos] = indices[pos]
         row_matval[pos] = row_entries[pos]

      # coefficient 1.0 of xi
      row_matind[numNonZeroElems] = self.numFeatures+example_idx
      row_matval[numNonZeroElems] = 1.0

      status = CPX.addrows (self.env, self.lp, numNewCols, numNewRows, numNonZeroElems+1, row_rhs,
                                  row_sense, row_matbeg, row_matind, row_matval)
      
      assert status == 0, 'Error ocurred while adding constraint.'

      #CPX.writeprob(self.env,self.lp,"problem_%d.SAV"%self.addedConstraints)
      #CPX.writeprob(self.env,self.lp,"problem_%d.LP"%self.addedConstraints)
      self.addedConstraints += 1

      return True
               
   def solve(self):
      # solve the problem
      CPX.qpopt(self.env,self.lp)

      # get the solution
      lpstat = CPX.getstat(self.env,self.lp)
      if lpstat in OPTIMAL or lpstat == 6:
         x, objval = CPX.getx(self.env,self.lp), CPX.getobjval(self.env,self.lp)
         print >> self.protocol, 'SIQP_CPX: Objective is: %f'%objval

         # Check that objective value is monotonically increasing
         #if math.fabs(objval - self.old_objective_value) >= 1e-4:
         #   assert objval >= self.old_objective_value

         self.old_objective_value = objval

         new_w = cb.matrix([0.0]*self.numVariables,(self.numVariables,1))
         for i in range(self.numVariables):
            new_w[i,0] = x[i]

         self.old_w = new_w

         # Check for non-negative slacks    
         sumOfSlacks = 0.0
         print >> self.protocol, 'Slacks are:'
         for i in range(self.numExamples):
            print >> self.protocol, '%f ' % new_w[self.numFeatures+i,0]
            assert new_w[self.numFeatures+i,0] >= 0.0
            sumOfSlacks += new_w[self.numFeatures+i,0]

         scalarProdDiff = (new_w[0:self.numFeatures,0].T * self.lastDelta)[0]
         # print >> self.protocol, 'Constraint: %f >= %d - %f'%(scalarProdDiff,self.lastLoss,new_w[self.numFeatures+self.lastIndex])

         # evalute parts of the objective
         regularizationPart = (0.5 *new_w.T*self.P*new_w)[0]
         lossPart = 1.0*self.C/self.numExamples * sumOfSlacks
         assert lossPart >= 0.0, 'Error we have a negative loss contribution'
         print >> self.protocol, 'Parts of objective: %e %e'%(regularizationPart,lossPart)

         return objval,new_w[0:self.numFeatures,0],new_w[self.numFeatures:,0]

      elif lpstat in UNBOUNDED: 
         print >> self.protocol, "Solution unbounded"
         assert False
      elif lpstat in INFEASIBLE: 
         print >> self.protocol, "Solution infeasible"
         assert False
      else: 
         print >> self.protocol, "Solution code: ", lpstat
         assert False

   def cpx_matrix(self,M):
      """
      This method creates a sparse CPLEX representation of a given
      cvxopt Matrix M.
      """

      nRows,nCols = M.size
      numEntries = nRows*nCols

      matbeg = IntMatrix([0]*nCols).data
      matcnt = IntMatrix([0]*nCols).data

      nonZeros = 0
      elemCounter = 0
      entry = 0.0

      valVec = [0.0]*numEntries
      indVec = [0.0]*numEntries

      for currentCol in range(nCols):
         nonZeros = 0
         matbeg[currentCol] = elemCounter

         for currentRow in range(nRows):
            entry = M[currentRow,currentCol]
            if math.fabs(entry) > self.EPSILON:
               nonZeros += 1
               indVec[elemCounter] = currentRow
               valVec[elemCounter] = entry
               elemCounter += 1

         matcnt[currentCol] = nonZeros

      totalNonZeros = elemCounter

      matind = IntMatrix([0]*totalNonZeros).data
      matval = FloatMatrix([0.0]*totalNonZeros).data

      for idx in range(totalNonZeros):
         matind[idx] = indVec[idx];
         matval[idx] = valVec[idx];
    
      print >> self.protocol, 'nRows,nCols: %d,%d'%(nRows,nCols)
      print >> self.protocol, 'totalNonZeros ' + str(totalNonZeros)

      return matbeg,matcnt,matind,matval

   def cpx_matrix_id(self):
      """
      This method creates a sparse CPLEX representation of a given
      cvxopt Matrix M.
      """

      nRows,nCols = self.numVariables,self.numVariables
      matbeg = IntMatrix([0]*nCols).data
      matcnt = IntMatrix([0]*nCols).data

      matind = IntMatrix([0]*nCols).data
      matval = FloatMatrix([0.0]*nCols).data

      for i in range(nCols):
         matbeg[i] = i
         matcnt[i] = 1
         matind[i] = i
         matval[i] = 1.0

      return matbeg,matcnt,matind,matval

   def delete(self):
      # free the problem
      CPX.freeprob(self.env,self.lp)
      # close CPLEX environment
      CPX.closeCPLEX(self.env)

def test():
   fh = open('siqp_cpx.log','w+')

   numFeatures = 3
   numExamples = 2
   s = SIQPSolver(numFeatures,numExamples,10.0,fh)

   d1 = cb.matrix([1,0,2],(numFeatures,1))
   d2 = cb.matrix([1,1,0],(numFeatures,1))

   s.addConstraint(d1,200.0,0,True)
   s.addConstraint(d2,105.0,1,True)

   w,slacks = s.solve()
   print w
   print slacks

   print w.T * d1 + slacks[0]
   print w.T * d2 + slacks[1]


   s.delete()
   del s
   fh.close()

if __name__ == '__main__':
   test()
