# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

from numpy import inf
from numpy.matlib import zeros
import pdb

# 3 * 30 supporting points: x_1 ... x_30 => 3 * 30 parameters (1 .. 90): y_1 ... y_30
# piecewise linear function:
# take score from SVM vektor (don_supp: case 1, acc_supp: case 2) and compute length of intron: case 3
# these are our values x
# 
#        | y_1                                          if x <= x_1
#        |
#        |        x_i+1 - x              x - x_i
# f(x) = | y_i * -----------  +  y_i+1 * -----------    if x_i <= x <= x_i+1
#        |       x_i+1 - x_i             x_i+1 - x_i
#        |
#        | y_30                                         if x_n <= x 
# 
#  y_i and y_i+1 parameters, so the fractions are saved in the weight vectors!

def calculateWeights(plf, scores):
   currentWeight = zeros((len(plf.limits),1))
   
   for k in range(len(scores)):
      value = scores[k]
      Lower = len([elem for elem in plf.limits if elem <= value])

      if Lower == 0:
         currentWeight[0] += 1
      elif Lower == len(plf.limits):
         currentWeight[-1] += 1
      else:
         # because we count from 0 in python
         Lower -= 1
         Upper = Lower+1 ; # x-werte bleiben fest
         #print value,Lower,Upper
         weightup  = 1.0*(value - plf.limits[Lower]) / (plf.limits[Upper] - plf.limits[Lower])
         weightlow = 1.0*(plf.limits[Upper] - value) / (plf.limits[Upper] - plf.limits[Lower])
         currentWeight[Upper] = currentWeight[Upper] + weightup
         currentWeight[Lower] = currentWeight[Lower] + weightlow

         #print plf.limits[Lower],plf.limits[Upper]
         #print weightup,weightlow,currentWeight[Upper],currentWeight[Lower]

   return currentWeight

def calculatePlif(plf, scores):
   plifed_scores=[] ;
   
   for k in range(len(scores)):
      value = scores[k]

      Lower = 0
      for elem in plf.limits:
          if elem<=value:
              Lower+=1 

      if Lower == 0:
          plifed_scores.append(plf.penalties[0])
      elif Lower == len(plf.limits):
          plifed_scores.append(plf.penalties[-1])
      else:
         Lower -= 1
         Upper = Lower+1 ;
         weightup  = 1.0*(value - plf.limits[Lower]) / (plf.limits[Upper] - plf.limits[Lower])
         weightlow = 1.0*(plf.limits[Upper] - value) / (plf.limits[Upper] - plf.limits[Lower])

         plifed_scores.append( plf.penalties[Upper]*weightup + plf.penalties[Lower]*weightlow )

   return plifed_scores 

def computeSpliceWeights(d, a, h, SpliceAlign, don_supp, acc_supp):
   ####################################################################################
   # 1. Donor: In don_supp stehen Werte der SVM., in SpliceAlign die Einsen
   ####################################################################################

   # Picke die Positionen raus, an denen eine Donorstelle ist
   try:
      DonorScores = [elem for pos,elem in enumerate(don_supp) if SpliceAlign[pos] == 1]
      if ( -inf in DonorScores ):
            print 'there are invalid donor scores, continue anyway.' ;
   except:
      print 'Error'
      #pdb.set_trace()

   #print 'donor'
   weightDon = calculateWeights(d,DonorScores)

   ####################################################################################
   # 2. Acceptor: In acc_supp stehen Werte der SVM., in SpliceAlign die Einsen
   ####################################################################################

   #Den Vektor Acceptorstellen durchgehen und die Gewichtsvektoren belasten:
   try:
      AcceptorScores = [elem for pos,elem in enumerate(acc_supp) if pos > 0 and SpliceAlign[pos] == 2]
      if ( -inf in AcceptorScores ):
            print 'there are invalid acceptor scores, continue anyway.' ;
   except:
      print 'Error'
      #pdb.set_trace()

   #print 'acceptor'
   weightAcc = calculateWeights(a,AcceptorScores)

   ####################################################################################
   # 3. Intron length: SpliceAlign: Gaps zaehlen und auf Gapgewichte addieren
   ####################################################################################

   intron_starts = []
   intron_ends   = []
   for pos,elem in enumerate(SpliceAlign):
      if elem == 1:
         intron_starts.append(pos)

      if elem == 2:
         intron_ends.append(pos)

   assert len(intron_starts) == len(intron_ends)

   for i in range(len(intron_starts)):
      assert intron_starts[i] < intron_ends[i]

   values = [0.0]*len(intron_starts)
   for pos in range(len(intron_starts)):
      values[pos] = intron_ends[pos] - intron_starts[pos] + 1

   #print 'introns found: ' + str(values)
   weightIntron = calculateWeights(h,values)

   return weightDon, weightAcc, weightIntron
