
"""
Functions for parsing palma parameter files,
to extract splice site predictors for qpalma.
"""

import bz2
import sys
import numpy
from numpy import mat,array,chararray,inf

class palma_model():
    """A container for palma parameters"""
    
    #acceptor
    acceptor_bins=None
    acceptor_limits=None
    acceptor_penalties=None

    acceptor_splice_b=None
    acceptor_splice_order=None
    acceptor_splice_window_left=None
    acceptor_splice_window_right=None
    acceptor_splice_alphas=None
    acceptor_splice_svs=None

    #donor
    donor_bins=None
    donor_limits=None
    donor_penalties=None
    
    donor_splice_b=None
    donor_splice_order=None
    #donor_splice_use_gc=None
    donor_splice_window_left=None
    donor_splice_window_right=None
    donor_splice_alphas=None
    donor_splice_svs=None

    # intron length
    intron_len_bins=None
    intron_len_limits=None
    intron_len_penalties=None
    intron_len_min=None
    intron_len_max=None
    intron_len_transform=None

    gene_len_max = None
    
    # substitution matrix
    substitution_matrix=None 



def parse_file(filename):
    if filename.endswith('.param.bz2'):
        fileptr=bz2.BZ2File(filename);
    else:
        sys.stderr.write('Expected palma parameter file with ending param.bz2\n')
    m=palma_model()

    sys.stdout.write("loading model file"); sys.stdout.flush()

    l=fileptr.readline();

    if l != '%palma definition file version: 1.0\n':
        sys.stderr.write("\nfile not a palma definition file\n")
        print l
        return None

    while l:
        if not ( l.startswith('%') or l.startswith('\n') ): # comment
            
            #acceptor
            if m.acceptor_bins is None: m.acceptor_bins=parse_value(l, 'acceptor_bins')
            if m.acceptor_limits is None: m.acceptor_limits=parse_vector(l, fileptr, 'acceptor_limits')
            if m.acceptor_penalties is None: m.acceptor_penalties=parse_vector(l, fileptr, 'acceptor_penalties') #DEAD

            if m.acceptor_splice_b is None: m.acceptor_splice_b=parse_value(l, 'acceptor_splice_b')
            if m.acceptor_splice_order is None: m.acceptor_splice_order=parse_value(l, 'acceptor_splice_order')
            if m.acceptor_splice_window_left is None: m.acceptor_splice_window_left=parse_value(l, 'acceptor_splice_window_left')
            if m.acceptor_splice_window_right is None: m.acceptor_splice_window_right=parse_value(l, 'acceptor_splice_window_right')
            if m.acceptor_splice_alphas is None: m.acceptor_splice_alphas=parse_vector(l, fileptr, 'acceptor_splice_alphas')
            if m.acceptor_splice_svs is None: m.acceptor_splice_svs=parse_string(l, fileptr, 'acceptor_splice_svs')

            #donor
            if m.donor_bins is None: m.donor_bins=parse_value(l, 'donor_bins')
            if m.donor_limits is None: m.donor_limits=parse_vector(l, fileptr, 'donor_limits')
            if m.donor_penalties is None: m.donor_penalties=parse_vector(l, fileptr, 'donor_penalties') #DEAD

            if m.donor_splice_b is None: m.donor_splice_b=parse_value(l, 'donor_splice_b')
            if m.donor_splice_order is None: m.donor_splice_order=parse_value(l, 'donor_splice_order')
            #if m.donor_splice_use_gc is None: m.donor_splice_use_gc=parse_value(l, 'donor_splice_use_gc')
            if m.donor_splice_window_left is None: m.donor_splice_window_left=parse_value(l, 'donor_splice_window_left')
            if m.donor_splice_window_right is None: m.donor_splice_window_right=parse_value(l, 'donor_splice_window_right')
            if m.donor_splice_alphas is None: m.donor_splice_alphas=parse_vector(l, fileptr, 'donor_splice_alphas')
            if m.donor_splice_svs is None: m.donor_splice_svs=parse_string(l, fileptr, 'donor_splice_svs')


            # intron length
            if m.intron_len_bins is None: m.intron_len_bins=parse_value(l, 'intron_len_bins')
            if m.intron_len_limits is None: m.intron_len_limits=parse_vector(l, fileptr, 'intron_len_limits')
            if m.intron_len_penalties is None: m.intron_len_penalties=parse_vector(l, fileptr, 'intron_len_penalties')
            if m.intron_len_min is None: m.intron_len_min=parse_value(l, 'intron_len_min')
            if m.intron_len_max is None: m.intron_len_max=parse_value(l, 'intron_len_max')
            if m.intron_len_transform is None: m.intron_len_transform=parse_value(l, 'intron_len_transform')

            # gene length
            if m.gene_len_max is None: m.gene_len_max=parse_value(l, 'gene_len_max')

            if m.substitution_matrix is None: m.substitution_matrix=parse_vector(l, fileptr, 'substitution_matrix')

        l=fileptr.readline()

    sys.stderr.write('done\n')
    return m

def parse_value(line, name):
    if (line.startswith(name)):
        #print 'found ' + name
        sys.stdout.write('.'); sys.stdout.flush()
        str = line[line.find('=')+1:-1] ;
        if str[0] == "'":
            return str[1:-1]
        else:
            #print float(str)
            return float(str)
    else:
        return None

def parse_vector(line, fileptr, name):
    mat = parse_matrix(line, fileptr, name)
    if mat is None:
     return mat
    else:
     mat = numpy.array(mat).flatten()
     return mat 

def parse_matrix(line, fileptr, name):
    if (line.startswith(name)):
        sys.stdout.write('.'); sys.stdout.flush()
        if line.find(']') < 0:
            l=''
            while l is not None and l.find(']') < 0:
                line+=l
                l=fileptr.readline()
            if l is not None and l.find(']') >= 0:
                line+=l

        if line.find(']') < 0:
            sys.stderr.write("matrix `" + name + "' ended without ']'\n")
            return None
        else:
            return mat(line[line.find('['):line.find(']')+1])
    else:
        return None

def parse_string(line, fileptr, name):
    if (line.startswith(name)):
        sys.stdout.write('.'); sys.stdout.flush()
        l=''
        lines=[]
        while l is not None and l.find(']') < 0:
            if l:
                lines+=[list(l[:-1])]
            l=fileptr.readline()

        if l.find(']') < 0:
            sys.stderr.write("string ended without ']'\n")
            return None
        else:
            #seqlen=len(lines[0])
            num=len(lines)
            #trdat = chararray((seqlen,num),1,order='FORTRAN')
            #for i in xrange(num):
            #    trdat[:,i]=lines[i]
            trdat = num*[[]]
            for idx,example in enumerate(lines):
                trdat[idx] = ''.join(example)
            return trdat
    else:
        return None

if __name__ == '__main__':
    #import hotshot, hotshot.stats

    def load():
        f='/fml/ag-raetsch/share/projects/palma/param_files/arabidopsis_ath1_ss=1_il=1.param.bz2'
        m=parse_file(f);

        print m.acceptor_bins is None
        print m.acceptor_limits is None
        print m.acceptor_penalties is None

        print m.acceptor_splice_b is None
        print m.acceptor_splice_order is None
        print m.acceptor_splice_window_left is None
        print m.acceptor_splice_window_right is None
        print m.acceptor_splice_alphas is None
        print m.acceptor_splice_svs is None

        print m.donor_bins is None
        print m.donor_limits is None
        print m.donor_penalties is None

        print m.donor_splice_b is None
        print m.donor_splice_order is None
        #print m.donor_splice_use_gc is None
        print m.donor_splice_window_left is None
        print m.donor_splice_window_right is None
        print m.donor_splice_alphas is None
        print m.donor_splice_svs is None
        print m.intron_len_bins is None
        print m.intron_len_limits is None
        print m.intron_len_penalties is None
        print m.intron_len_min is None
        print m.intron_len_max is None
        print m.intron_len_transform is None

        print m.substitution_matrix is None

        print 'intron_len_transform'
        print m.intron_len_transform
        print 'substitution_matrix'
        print m.substitution_matrix

    load()

    #prof = hotshot.Profile("model.prof")
    #benchtime = prof.runcall(load)
    #prof.close()
    #stats = hotshot.stats.load("model.prof")
    #stats.strip_dirs()
    #stats.sort_stats('time', 'calls')
    #stats.print_stats(20)
