#!/usr/bin/sample.py

import pickle
import random
import sys
import os
import re
import pdb

def parse_options(argv):

    """Parses options from the command line """

    from optparse import OptionParser, OptionGroup

    parser = OptionParser()
    required = OptionGroup(parser, 'REQUIRED')
    required.add_option('-q', '--qualities', dest='qualities', metavar='FILE', help='read qualities, e.g., generated by create_quality_sample.sh', default='-')
    required.add_option('-f', '--fasta', dest='fasta', metavar='FILE', help='file storing the reads created with parse_reads_from_gff.py', default='-')
    required.add_option('-m', '--mismatch', dest='mismatch', metavar='FILE', help='the output of mismatch.py according to which reads are \
                        mutated based on nucleotide composition and qualities', default='-')
    required.add_option('-l', '--logfile', dest='logfile', metavar='FILE', help='file created while sampling reads to reads.fa with parse_reads_from_gff.py', default='-')
    required.add_option('-o', '--optimal', dest='optimal', metavar='FILE', help='will store the optimal alignment of reads in fasta to the genome in genome.config', default='-')
    optional = OptionGroup(parser, 'OPTIONAL')
    optional.add_option('-s', '--simple', dest='simple', action='store_true', help='does not create optimal alignment, cannot simulate deletions', default=False)
    optional.add_option('-p', '--paired', dest='paired', action='store_true', help='generates paired optimal alignments, needs paired input reads', default=False)
    optional.add_option('-g', '--generrors', dest='generrors', action='store_false', help='if set, NO mutations are generated [ default: unset ]', default=True)
    required.add_option('-O', '--opt_sam', dest='opt_sam', metavar='FILE', help='will store the optimal alignment additionally in SAM format', default='-')
    optional.add_option('-e', '--errorinfo', dest='errorinfo', action='store_true', help='show error info in created read [ default: not set ]', default=False)
    optional.add_option('-F', '--force_deletions', dest='force_deletions', action='store_true', help='forces deletions in simple mode [ default: not set ]', default=False)
    optional.add_option('-i', '--insertprob', dest='p_insert', metavar='FLOAT', help='probability for an insertion during a mutation [ default: 0.05 ]', type='float', default=0.05)
    optional.add_option('-n', '--noise', dest='noise', metavar='FLOAT', help='probability that an unmutated position will be mutated nevertheless [ default: 0.0 ]', type='float', default=0.0)
    parser.add_option_group(required)
    parser.add_option_group(optional)

    (options, args) = parser.parse_args()
    
    if (options.simple and len(argv) < 4) or (not options.simple and (len(argv) < 6 or options.p_insert > 1.0 or options.noise > 1.0 or '-' in [options.fasta, options.mismatch, options.logfile, options.optimal])):
        parser.print_help()
        sys.exit(2)

    return options

def handle_insertions(logread, sizes, pos):
    """Adapts block sizes and block coordinates corresponding to insertions in the read. If a position is inserted, the LAST block becomes smaller!"""

    re_start = re.search('\[\-.\]', logread)
    while re_start != None:
        subst = len( re.findall('\[', logread[:re_start.start()]) )
        start = re_start.start() - (3 * subst)
        cum_size = 0
        for i in range(len(sizes)):
            cum_size += sizes[i]
            if start < cum_size:
                ## Do not consider insertions for position-calculations
                #if start == 0:
                #    pos[i] -= 1
                #elif start == cum_size - sizes[i]:
                #   ## pos[i] += 1
                #   pos[i] -= 1
                if  sizes[-1] == 0:
                    sizes = sizes[:-1]
                    pos = pos[:-1]
                sizes[-1] -= 1
                #sizes[i] -= 1
                break
        logread = logread.replace(logread[re_start.start():re_start.start() + 4], 'X', 1)
        re_start = re.search('\[\-.\]', logread)
       
        if  sizes[-1] == 0:
            sizes = sizes[:-1]
            pos = pos[:-1]
 
def handle_deletions(logread, sizes, pos):
    """Adapts block sizes and block coordinates corresponding to deletions at the block's ending or 
    the block's beginning. Deletions inside a block do NOT change alignment coordinates! """

    re_start = re.search('\[.\-\]', logread)
    while re_start != None:
        subst = len( re.findall('\[', logread[:re_start.start()]) )
        start = re_start.start() - (3 * subst)
        cum_size = 0
        for i in range(len(sizes)):
            cum_size += sizes[i]
            if start < cum_size:
                if start == cum_size - sizes[i]:
                    pos[i] += 1
                    sizes[i] -= 1
                elif start == cum_size - 1:
                    sizes[i] -= 1
                sizes[-1] += 1
                #else:
                #    sizes[i] += 1
                break
        logread = logread.replace(logread[re_start.start():re_start.start() + 4], 'X', 1)
        re_start = re.search('\[.\-\]', logread)
        

def main(options):
    
    # print >> sys.stderr, "Producing sample using\n ... quality file", options.qualities, "\n ... fasta file", \
    # options.fasta, "\n ... mismatch file", options.mismatch, "\n reading log from:", options.logfile, "\n writing optimal alignment to", options.optimal
    
    ((qualities, substitutions), dummy) = pickle.load(file(options.mismatch))

    if not (options.qualities == '-'):
        qfile = file(options.qualities)
        
    if options.paired:
        _ffile = options.fasta.strip().split(',')
        if len(_ffile) != 2:
            print >> sys.stderr, "\nERROR: Exactly two fasta files containing paired reads are expected as input: -f FILE1,FILE2\n"
            sys.exit(2)
        ffile = file(_ffile[0])
        ffile_2 = file(_ffile[1])
    else:
        ffile = file(options.fasta)

    if options.simple and not options.force_deletions:
        log = True
        sample_source = ('A', 'C', 'G', 'T', 'N')
    else:
        if options.paired:
            _log = options.logfile.strip().split(',')
            if len(_log) != 2:
                print >> sys.stderr, "\nERROR: Exactly two log files are expected as input: -l LOG1,LOG2\n"
                sys.exit(2)
            log = open(_log[0], 'r')
            log_2 = open(_log[1], 'r')
        else:
            log = open(options.logfile, 'r')
        sample_source = ('A', 'C', 'G', 'T', 'N', '-')
        
    opt = open(options.optimal, 'w')
    if options.opt_sam != '-':
        opt_sam = open(options.opt_sam, 'w')

    fh = ffile.readline()
    if options.paired:
        fh_2 = ffile_2.readline()
    fl = ffile.readline()
    if options.paired:
        fl_2 = ffile_2.readline()
   
    copy_fl = fl
    initial_len = len(copy_fl.strip())
    del copy_fl

    if (options.qualities == '-'):
        qlist = ['h' * len(fl)] 
    else:    
        ql = qfile.readline()
        qlist = [ql]
        while ql:
            ql = qfile.readline()
            if len(ql)>10:
                qlist.append(ql)

    if not options.simple:
        logline = log.readline().strip().split('\t')
        if options.paired:
            logline_2 = log_2.readline().strip().split('\t')
 
    #random.seed(17)
    random.seed(os.urandom(128))

    process_second = False

    while fh and fl and log and (not options.paired or (fh_2 and fl_2 and log_2)):
        ql = qlist[random.randint(0, len(qlist)-1)]
        while len(ql) < len(fl):
            ql = ql[:(len(ql)) / 2 + 1] + ql[len(ql) / 2:]
#        if extended:
#            print >> sys.stderr, 'warning! - had to extend quality line \n%s to \n%s ' % (old_line, ql)
        if len(ql) > len(fl):
            ql = ql[:len(fl)]
        print '@' + fh[1:].strip()
        read = ''
        logread = ''
        qual = '' 
        comp_del = 0

        if not options.simple:
            if len(logline) < 2:
                break
        ql.strip()
        fl.strip()

##        while j < len(ql) - 1:
        for i in xrange(len(ql) - 1):
            
            q = ql[i]
            f = fl[i].upper()
            try:
                (a, b) = qualities[(f, q)]
            except KeyError:
                a = 0
                b = 1
            if options.generrors and random.uniform(0, 1) < (float(a)/float(b) * (1.0 - options.noise)) + options.noise and random.uniform(0, 1) < options.p_insert:
                cum = 0
                p = random.uniform(0, 1)
                denom = 0
                for s in ('A', 'C', 'G', 'T', 'N', '-'):
                    try:
                        denom += substitutions[('-', s)]
                    except KeyError:
                        pass

                for s in ('A', 'C', 'G', 'T', 'N', '-'):
                    try:
                        cum += substitutions[('-', s)]
                    except KeyError:
                        pass
                    #print "p", p, "cum", cum/float(denom)
                    if denom > 0 and p < cum / float(denom):
                        if s == '-':
                            c = ''
                            lr = ''
                        else:
                            if options.errorinfo == 1:
                                c = '[-' + s + ']'
                            else:
                                c = s
                            lr = '[-' + s + ']'
                        break
                read += c
                qual += ql[i] 
                logread += lr

            q = ql[i]
            f = fl[i].upper()
            try:
                (a, b) = qualities[(f, q)]
            except KeyError:
                a = 0
                b = 1
            if options.generrors and random.uniform(0, 1) < (float(a) / float(b) * (1.0 - options.noise)) + options.noise :
                cum = 0
                p = random.uniform(0, 1)
                denom = 0

                for s in sample_source:
                    try:
                        denom += substitutions[(f, s)]
                    except KeyError:
                        pass

                for s in sample_source:
                    try:
                        cum += substitutions[(f, s)]
                    except KeyError:
                        pass
                    #print "p", p, "cum", cum/float(denom)
                    if denom > 0 and p < cum / float(denom):
                        if s == '-':
                            if options.errorinfo == 1:
                                c = '[' + f + '-]'
                            else:
                                c = ''
                            q = '' 
                            lr = '[' + f + '-]'
                            # compensate the deletion by adding additional read to the end
                            comp_del += 1
                        else:
                            if options.errorinfo == 1:
                                c = '[' + f + s + ']'
                            else:
                                c = s
                            lr = '[' + f + s + ']'
                        break
            else:
                c = f
                lr = f
            
            read += c
            qual += q 
            logread += lr

        log_with_errorinfo = logread

        # handle insertions
        while len(read) > initial_len:
            if options.errorinfo == 1 and read[-1] == ']':
                read = read[:-4]
            else:
                if re.search('\[.-\]', logread[-4:]) == None:
                    read = read[:-1]
                    qual = qual[:-1]
            if logread[-1] == ']':
                if re.search('\[.-\]', logread[-4:]) != None:
                    comp_del -= 1
                logread = logread[:-4]
            else:
                logread = logread[:-1]

        if not options.simple:
            # handle deletions
            if len(read) < initial_len:
                gap = initial_len - len(read)
                read += logline[7].strip()[:gap].upper()
                logread += logline[7].strip()[:gap].upper()
                qual += gap*qual[-1]

        # print reads in fastq format 
        print read.strip()
        print '+' + fh[1:].strip()
        print qual.strip()

        if not options.simple:
            # create optimal alignment
            pos = logline[5][1:-1].split(',')
            pos = [int(i) for i in pos]
            sizes = logline[6][1:-1].split(',')
            sizes = [int(i) for i in sizes]

            cloneread = logread

            # adapt block sizes in case of deletions at the beginning or ending of a block
            # MUST handle deletions first!
            handle_deletions(logread, sizes, pos)
            handle_insertions(logread, sizes, pos)
            if comp_del > 0 and int(logline[8].strip()) > 0:
                pos.append(int(logline[8].strip()))
                sizes.append(comp_del)

            # write optimal alignment / training sample in BED format
            tempstr = ''
            tempstr += (logline[2] + '\t' + str(pos[0]) +  '\t' + str(pos[-1] + sizes[-1]) + '\t')
            tempstr += (logline[0] + '\t' + str(len(read) - len(re.findall('\[', cloneread))) + '\t+\t0\t0\t0,0,0\t')
            tempstr += (str(len(sizes)) + '\t')
            for s in sizes:
                tempstr += (str(s) + ',')
            tempstr = tempstr[:-1]
            tempstr += '\t'
            # print relative block starts
            for p in pos:
                tempstr += (str(p - pos[0]) + ',')
            tempstr = tempstr[:-1]
            tempstr += '\t' + cloneread.strip() + '\t' + qual + '\t'
            for i in range(len(pos)):
                tempstr += (str(pos[i]) + '\t' + str(pos[i] + sizes[i]) + '\t')
            tempstr = tempstr[:-1]
            print >> opt, tempstr

            ## write optimal alignment in SAM format
            if options.opt_sam != '-':
                tempstr = ''
                # read_id
                tempstr += (logline[0] + '\t')
                # flag
                if options.paired:
                    _flag = 3 # paired in sequencing and mapped as pair
                else:
                    _flag = 0
                if len(logline) >= 10:
                    if logline[9] == '-':
                        _flag += 16
                    else:
                        _flag += 32
                if options.paired:
                    if logline[0][-1] == '1':   
                        _flag += 64
                    else:
                        _flag += 128
                tempstr += (str(_flag) + '\t')
                # chromosome
                tempstr += (logline[2] + '\t')
                # start pos
                if logline[9] == '-':
                    tempstr += (str(int(logline[11]) + 1) + '\t')
                else:
                    tempstr += (str(int(logline[5][1:-1].strip().split(',')[0]) + 1) + '\t')
                # mapping quality
                tempstr += '1\t'
                # CIGAR
                _cigar = ''
                for p in range(len(log_with_errorinfo)):
                    if log_with_errorinfo[p] != '[':
                        _cigar += 'M'
                    else:
                        if log_with_errorinfo[p + 1] == '-':
                            _cigar += 'I'
                        elif log_with_errorinfo[p + 2] == '-':
                            _cigar += 'D'
                        else:
                            _cigar += 'M'
                        p += 3
                _c = 0
                __cigar = ''
                _last = _cigar[0]
                _cum_sizes = [sizes[0]]
                for p in range(1, len(sizes), 1):
                    _cum_sizes.append(_cum_sizes[p - 1] + sizes[p])
                _idx = 0
                for p in range(len(_cigar)):
                    if _cigar[p] == _last:
                        _c += 1
                    else:
                        __cigar += (str(_c) + _last)
                        _c = 1
                    _last = _cigar[p]
                    if p < (len(_cigar) - 1) and p == (_cum_sizes[_idx] - 1):
                        __cigar += (str(_c) + _last)
                        _c = 0
                        __cigar += (str(pos[_idx + 1] - pos[_idx] - sizes[_idx]) + 'N')
                        _idx += 1
                tempstr += (__cigar + str(_c) + _last + '\t')
                # MateRefName
                tempstr += '=\t'
                # Mate start pos
                if options.paired:
                    if logline[9] == '-':
                        tempstr += (str(int(logline_2[5][1:-1].strip().split(',')[0]) + 1) + '\t')
                    else:
                        tempstr += (str(int(logline_2[11]) + 1) + '\t')
                else:
                    tempstr += '*\t'
                # insert size
                tempstr += '*\t'
                # sequence
                tempstr += (read.strip() + '\t')
                # quality
                tempstr += (qual.strip() + '\t')


                print >> opt_sam, tempstr

        process_second = not process_second
        if options.paired and process_second:
            (fh, fh_2) = (fh_2, fh)
            (fl, fl_2) = (fl_2, fl)
            (logline, logline_2) = (logline_2, logline)
        else:
            fh = ffile.readline()
            if options.paired:
                fh_2 = ffile_2.readline()
            fl = ffile.readline()
            if options.paired:
                fl_2 = ffile_2.readline()
            if not options.simple:
                logline = log.readline().strip().split('\t')
                if options.paired:
                    logline_2 = log_2.readline().strip().split('\t')
    opt.close()
    if options.opt_sam != '-':
        opt_sam.close()

if __name__ == '__main__':
    main(parse_options(sys.argv))
